@@ -42,16 +42,19 @@ extern int errno;
 #include "util.h"
 
 /* The save_state gives a description of what we are doing */
-static enum {
+typedef enum {
   MANAGER_IDLE,
   SENDING_MESSAGES,
   SENDING_INTERACT,
+  SENDING_INTERACT_2,
   SAVE_PHASE_1,
   SAVE_PHASE_2,
   SAVE_CANCELLED,
   STARTING_SESSION,
   SHUTDOWN
-} save_state;
+} SaveState;
+
+SaveState save_state;
 
 /* List of clients which have been started but have yet to register */
 GSList *pending_list = NULL;
@@ -265,7 +268,7 @@ free_client (Client *client)
 {
  GSList *list; 
 
- gsm_verbose ("free_client(): %p\n", client);
+ gsm_verbose ("free_client(): %p (%s)\n", client, client->id ? client->id : "(null)");
 
  if (! client)
     return;
@@ -315,17 +318,22 @@ free_client (Client *client)
   g_slist_free (client->get_prop_replies);
   client->get_prop_replies = NULL; /* sanity */
 
+  if (client->last_discard)
+    SmFreeProperty (client->last_discard);
+
   g_free (client);
 }
 
 /* Run a command on a client. */
 gint
-run_command (Client* client, const gchar* command)
+run_command_prop (Client     *client, 
+		  const char *command, 
+		  SmProp     *prop)
 {
   int argc, envc, envpc, pid = -1;
   char **argv, *dir, **envv, **envp, *restart_info;
 
-  if (find_vector_property (client, command, &argc, &argv))
+  if (gsm_parse_vector_prop (prop, &argc, &argv))
     {
       if (!find_string_property (client, SmCurrentDirectory, &dir))
 	dir = NULL;
@@ -387,6 +395,19 @@ run_command (Client* client, const gchar* command)
   return pid;
 }
 
+int
+run_command (Client      *client,
+	     const gchar *command)
+{
+  SmProp *prop;
+
+  prop = find_property_by_name (client, command);
+  if (!prop)
+    return -1;
+
+  return run_command_prop (client, command, prop);
+}
+
 /* Send a message to every client on LIST.  */
 static void
 send_message (GSList **list, message_func *message)
@@ -503,19 +524,23 @@ purge (gpointer data)
 }
 
 static void
-reincarnate_client (Client *old_client, Client *new_client)
+reincarnate_client (Client *old_client,
+		    Client *new_client)
 {
-  new_client->properties        = old_client->properties;
-  new_client->priority          = old_client->priority;
-  new_client->match_rule        = old_client->match_rule;
-  new_client->attempts          = old_client->attempts;
-  new_client->connect_time      = old_client->connect_time;
-  new_client->command_data      = old_client->command_data;
+  new_client->properties    = old_client->properties;
+  new_client->priority      = old_client->priority;
+  new_client->match_rule    = old_client->match_rule;
+  new_client->attempts      = old_client->attempts;
+  new_client->connect_time  = old_client->connect_time;
+  new_client->command_data  = old_client->command_data;
+  new_client->session_saved = old_client->session_saved;
   
   old_client->properties = NULL;
+
   if (old_client->handle)
     new_client->handle = command_handle_reassign (old_client->handle, new_client);
   old_client->handle = NULL;
+
   free_client (old_client);	  
 }
 
@@ -810,7 +835,8 @@ interact_request (SmsConn connection, SmPointer data, int dialog_type)
   /* These seem to be the only circumstances in which interactions
      can be allowed without breaking the protocol.
      FIXME: Should we inform the user of bogus requests ? */
-  if (save_state == SAVE_PHASE_1 &&
+  if ((save_state == SAVE_PHASE_1 ||
+       save_state == SAVE_PHASE_2) &&
       (g_slist_find (save_yourself_list, client) ||
        g_slist_find (save_yourself_p2_list, client)))
     {
@@ -821,10 +847,14 @@ interact_request (SmsConn connection, SmPointer data, int dialog_type)
 					     compare_interact_request);
       if (send) 
 	{
-	  save_state = SENDING_INTERACT;
+	  save_state = (save_state == SAVE_PHASE_1) 
+	    ? SENDING_INTERACT
+	    : SENDING_INTERACT_2;
 	  interact_ping_replied = TRUE;
 	  SmsInteract (connection);
-	  save_state = SAVE_PHASE_1;
+	  save_state = (save_state == SENDING_INTERACT)
+	    ? SAVE_PHASE_1
+	    : SAVE_PHASE_2;
 	}
     }
 }
@@ -965,11 +995,15 @@ interact_done (SmsConn connection, SmPointer data, gboolean cancel)
 
   if (interact_list)
     {
-      save_state = SENDING_INTERACT;
+      save_state = (save_state == SAVE_PHASE_1)
+	? SENDING_INTERACT
+	: SENDING_INTERACT_2;
       client = (Client *) interact_list->data;
       interact_ping_replied = TRUE;
       SmsInteract (client->connection);
-      save_state = SAVE_PHASE_1;
+      save_state = (save_state == SENDING_INTERACT)
+	? SAVE_PHASE_1
+	: SAVE_PHASE_2;
     }
   /* Check in case the client sent a SaveYourselfDone during
      the interaction */ 
@@ -1006,9 +1040,14 @@ process_save_request (Client* client, int save_type, gboolean shutdown,
 	  while (live_list) 
 	    {
 	      Client *tmp_client = (Client *) live_list->data;
+
 	      client_event (tmp_client->handle, GsmSave);
 	      REMOVE (live_list, tmp_client);
 	      APPEND (save_yourself_list, tmp_client);
+
+	      tmp_client->last_discard = gsm_prop_copy (
+			find_property_by_name (tmp_client, SmDiscardCommand));
+
 	      SmsSaveYourself (tmp_client->connection, save_type, shutting_down,
 			       interact_style, fast);
 	    }
@@ -1022,6 +1061,10 @@ process_save_request (Client* client, int save_type, gboolean shutdown,
       client_event (client->handle, GsmSave);
       REMOVE (live_list, client);
       APPEND (save_yourself_list, client);
+
+      client->last_discard = gsm_prop_copy (
+		find_property_by_name (client, SmDiscardCommand));
+
       SmsSaveYourself (client->connection, 
 		       save_type, 0, interact_style, fast);
     }
@@ -1163,7 +1206,7 @@ update_save_state (void)
       if (shutting_down)
 	{
 	  Client *warner = get_warner();
-	  GSList* list;
+	  GSList *list;
 
 	  while (purge_drop_list)
 	    {
@@ -1173,11 +1216,13 @@ update_save_state (void)
 	      REMOVE (purge_drop_list, client);
 	      free_client (client);
 	    }
+
 	  maybe_write_session ();
-	    
+
 	  for (list = save_finished_list; list;)
 	    {
-	      Client *client = (Client *)list->data;
+	      Client *client = list->data;
+
 	      list = list->next;
 	      if (client != warner)
 		SmsDie (client->connection);
@@ -1260,6 +1305,33 @@ update_save_state (void)
     }
 }
 
+/*
+ * The basic rule here is that every time we
+ * 'forget' a RestartCommand we need to run
+ * the DiscardCommand.
+ */
+static void
+maybe_run_discard_commands (Client *client)
+{
+  SmProp *discard;
+
+  discard = find_property_by_name (client, SmDiscardCommand);
+
+  /* Broken Client: not saving its session data on a
+   * per-SaveYourself basis.
+   * See 'Advice to Implementors' under the SaveYourself
+   * section in the XSMP spec.
+   */
+   if (gsm_prop_compare (client->last_discard, discard))
+     return;
+
+   if (client->last_discard && !client->session_saved)
+     run_command_prop (client, SmDiscardCommand, client->last_discard);
+
+   if (discard && shutting_down && !(autosave || save_selected))
+     run_command_prop (client, SmDiscardCommand, discard);
+}
+
 static void
 save_yourself_done (SmsConn connection, SmPointer data, gboolean success)
 {
@@ -1349,6 +1421,8 @@ save_yourself_done (SmsConn connection, SmPointer data, gboolean success)
       set_properties(client->connection, (SmPointer)client, nprops, props);
     }
 
+  maybe_run_discard_commands (client);
+
   REMOVE (save_yourself_list, client);
   REMOVE (save_yourself_p2_list, client);
   APPEND (save_finished_list, client);      
@@ -1452,7 +1526,7 @@ client_clean_up (Client* client)
       break;
     }
   
-  if (save_state == SENDING_INTERACT)
+  if (save_state == SENDING_INTERACT || save_state == SENDING_INTERACT_2)
     {
       if (interact_list)
 	{
@@ -1460,7 +1534,9 @@ client_clean_up (Client* client)
 	  interact_ping_replied = TRUE;
 	  SmsInteract (client->connection);
 	}
-      save_state = SAVE_PHASE_1;
+      save_state = (save_state == SENDING_INTERACT)
+	? SAVE_PHASE_1
+	: SAVE_PHASE_2;
     }
   
   update_save_state ();
